/* 2d marching cubes algorithm by Scott Klasky */

#include <stdio.h>
#include <math.h>

/* Need to make a structure that I can then acces the grid functions */

typedef struct{
	int *level;
	double **data;
}GF;

	static void print_uij(int nx,int ny, double x[], double y[],
               double **val,char *fname) { 
		FILE *fp;
		int i,j;
		fp = fopen(fname,"w");
      j = (ny-1)/2;
      for (i=0;i<nx;i++) {
			fprintf(fp,"%lf %lf\n",x[i],val[j][i]);
      }
   }



   static void set_up_u(int nx, int ny, double x[], double y[],
                      double **u, double **uxct, double h){
	int i,j;
      for ( j=0;j<ny;j++) {
         for (i=0;i<nx;i++) {
            uxct[j][i] = 1.0/sqrt(x[i]*x[i]+y[j]*y[j]+0.5);
            u[j][i] = 0.0;
         }
      }
      for ( j=0;j<ny;j++) {
         u[j][0] = uxct[j][0];
         u[j][nx-1] = uxct[j][nx-1];
      }
      for ( i=0;i<nx;i++) {
         u[0][i] = uxct[0][i];
         u[ny-1][i] = uxct[ny-1][i];
      }

   }


   static void set_up_rhs(int nx,int ny,double x[],double y[],double **rhs){
	int i,j;
      for (j=0;j<ny;j++) {
         for (i=0;i<nx;i++) {
            rhs[j][i] = 32.0*(x[i]*x[i]+y[j]*y[j]-1.0)/sqrt(
                        (4*x[i]*x[i]+4*y[j]*y[j]+2.0)*
                        (4*x[i]*x[i]+4*y[j]*y[j]+2.0)*
                        (4*x[i]*x[i]+4*y[j]*y[j]+2.0)*
                        (4*x[i]*x[i]+4*y[j]*y[j]+2.0)*
                        (4*x[i]*x[i]+4*y[j]*y[j]+2.0));
         }
      }
   }

   double relaxc(int nx, int ny, double **u,double **rhs,
                 double **res, double h){

		int isw,jsw,i,j,pass;
		double sum2,sum3;
      double h2    = 1.0/( h * h);
      double jacel = (h * h)/ 4.0;
      double sum = 0.0;

      jsw = 0;
      isw;
      for (pass=1;pass<3;pass++){
         isw = jsw;
         for (j=1;j<ny-1;j++) {
            for (i=isw+1;i<nx-1;i+=2) {

               res[j][i] = h2 * ( u[j+1][i] + u[j-1][i] +
                                 u[j][i+1] + u[j][i-1] -
                                 4.0 * u[j][i] )  - rhs[j][i];
  
               u[j][i] += jacel * res[j][i];
               sum = sum + res[j][i] * res[j][i] ;

            }
         isw=1 - isw;
         }
         jsw = 1 - jsw;
      }
      sum2 = sum / nx / nx / ny / ny;
      sum3 = sqrt(sum2);

      return sum3;
   }


   double erroru(int nx, int ny, double **u, double **ux){
      double error,val;
		int i,j;
      error = 0.0;
      for (j=0;j<ny;j++){
         for (i=0;i<nx;i++){
            val = u[j][i]-ux[j][i];
            error = error + val * val;
         }
      }
      error = error / nx /ny;
      return error;
   }





/* The main program */
main( int argc, char *argv[]) {
int nxf,nyf,level;
int ncycle,pre,pst;
int i,j;
int prec;
double l2norm,l2err;
double **u,**uxct,**rhs,**tau,**temp;
double *x,*y;
double h;
	if (argc < 2) {
		printf("Usage mg2d level ncycles pre pst\n");
		exit();
	}

	level   = atoi(argv[1]);
	ncycle  = atoi(argv[2]);
	pre     = atoi(argv[3]);
	pst     = atoi(argv[4]);


	nxf = 1+ pow(2,level);
	nyf = nxf;

	printf("nxf= %d  \n",nxf);
	printf("level= %d \n",level);

/* now allocate the memory for the grid functions */

	u = (double **) malloc(nyf*sizeof(double));
	uxct = (double **) malloc(nyf*sizeof(double));
	rhs = (double **) malloc(nyf*sizeof(double));
	tau = (double **) malloc(nyf*sizeof(double));
	temp = (double **) malloc(nyf*sizeof(double));
	x = (double *)malloc(sizeof(double)*nxf);
	y = (double *)malloc(sizeof(double)*nyf);

   for (i=0;i<nxf;i++) {
		u[i]    = (double *)malloc(sizeof(double)*nxf);
		uxct[i] = (double *)malloc(sizeof(double)*nxf);
		rhs[i] = (double *)malloc(sizeof(double)*nxf);
		tau[i]  = (double *)malloc(sizeof(double)*nxf);
		temp[i]  = (double *)malloc(sizeof(double)*nxf);
	}

   h = 2.0/(nxf-1);

   for (i=0;i<nxf;i++) {
      x[i] = -1.0 + (i)*h;
   }
   for (i=0;i<nyf;i++) {
      y[i] = -1.0 + (i)*h;
   }

			
    set_up_u(nxf, nyf, x, y,u,uxct,h);
    print_uij(nxf,nyf, x, y,uxct,"uxct.dat");
    set_up_rhs(nxf, nyf, x, y,rhs);
    print_uij(nxf,nyf, x, y,rhs,"rhs.dat");

/* now lets try to relax........ */

	for (prec=0;prec<pre;prec++) {
		l2norm = relaxc(nxf,nyf,u,rhs,temp,h);
		l2err = erroru(nxf,nyf,u,uxct);
		printf("%d %e %e\n",prec,l2norm,l2err);
	}

}
